from enum import StrEnum


class CSIPAusVersion(StrEnum):
    """The various version identifiers for CSIP-Aus. Used for distinguishing what tests are compatible with what
    released versions CSIP-Aus."""

    RELEASE_1_2 = "v1.2"
    BETA_1_3_STORAGE = "v1.3-beta/storage"


class CSIPAusResource(StrEnum):
    """Labels for each resource type that the server/client tests might reference. This is not designed to be
    an exhaustive list of all SEP2 / CSIP-Aus models - only the entities that might have tests applied to them."""

    DeviceCapability = "DeviceCapability"

    Time = "Time"

    MirrorUsagePointList = "MirrorUsagePointList"
    MirrorUsagePoint = "MirrorUsagePoint"

    EndDeviceList = "EndDeviceList"
    EndDevice = "EndDevice"
    ConnectionPoint = "ConnectionPoint"
    Registration = "Registration"

    FunctionSetAssignmentsList = "FunctionSetAssignmentsList"
    FunctionSetAssignments = "FunctionSetAssignments"

    DERProgramList = "DERProgramList"
    DERProgram = "DERProgram"

    DERControlList = "DERControlList"
    DERControl = "DERControl"

    DefaultDERControl = "DefaultDERControl"

    DERList = "DERList"
    DER = "DER"
    DERCapability = "DERCapability"
    DERSettings = "DERSettings"
    DERStatus = "DERStatus"

    SubscriptionList = "SubscriptionList"
    Subscription = "Subscription"

    Notification = "Notification"  # A Notification isn't normally discoverable - it's received via pub/sub webhook


class CSIPAusReadingLocation(StrEnum):
    Site = "Site"  # The reading is measured at the site's connection point
    Device = "Device"  # The reading is measured at the actual device (behind the meter)


class CSIPAusReadingType(StrEnum):
    """A non exhaustive set of CSIPAus reading types / role flags that can be specified in tests"""

    ActivePowerAverage = "ActivePowerAverage"
    ActivePowerInstantaneous = "ActivePowerInstantaneous"
    ActivePowerMaximum = "ActivePowerMaximum"
    ActivePowerMinimum = "ActivePowerMinimum"

    ReactivePowerAverage = "ReactivePowerAverage"
    ReactivePowerInstantaneous = "ReactivePowerInstantaneous"
    ReactivePowerMaximum = "ReactivePowerMaximum"
    ReactivePowerMinimum = "ReactivePowerMinimum"

    FrequencyAverage = "FrequencyAverage"
    FrequencyInstantaneous = "FrequencyInstantaneous"
    FrequencyMaximum = "FrequencyMaximum"
    FrequencyMinimum = "FrequencyMinimum"

    VoltageSinglePhaseAverage = "VoltageSinglePhaseAverage"
    VoltageSinglePhaseInstantaneous = "VoltageSinglePhaseInstantaneous"
    VoltageSinglePhaseMaximum = "VoltageSinglePhaseMaximum"
    VoltageSinglePhaseMinimum = "VoltageSinglePhaseMinimum"


def is_list_resource(resource: CSIPAusResource) -> bool:
    """Returns true if the specified resource is classified as a list resource (i.e. it supports list query params) and
    will return an entity with list attributes (eg 'all')"""
    return resource.name.endswith("List")  # This is a really simple method but it should work



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import IntEnum, auto
from typing import Any

from cactus_test_definitions.csipaus import (
    CSIPAusReadingLocation,
    CSIPAusReadingType,
    CSIPAusResource,
)
from cactus_test_definitions.errors import TestProcedureDefinitionError
from cactus_test_definitions.variable_expressions import is_resolvable_variable


class ParameterType(IntEnum):
    """The various "basic" types that can be set in YAML for a parameter. Rudimentary type checking
    will try to enforce these but they can't be guaranteed"""

    String = auto()
    Integer = auto()
    Float = auto()
    Boolean = auto()
    DateTime = auto()  # TZ aware datetime
    ListString = auto()  # List of strings
    HexBinary = auto()
    CSIPAusResource = auto()  # Member of cactus_test_definitions.csipaus.CSIPAusResource
    ListCSIPAusResource = auto()  # List where each member is a cactus_test_definitions.csipaus.CSIPAusResource
    CSIPAusReadingType = auto()  # Member of cactus_test_definitions.csipaus.CSIPAusReadingType
    ListCSIPAusReadingType = auto()  # List where each member is a cactus_test_definitions.csipaus.CSIPAusReadingType
    CSIPAusReadingLocation = auto()  # Member of cactus_test_definitions.csipaus.CSIPAusReadingLocation
    ReadingTypeValues = auto()  # A dict of type dict[CSIPAusReadingType, list[float]], each list has the same length
    UnsignedInteger = auto()


@dataclass(frozen=True)
class ParameterSchema:
    """What parameters can be passed to a given action/check. Describes a single optional/mandatory field"""

    mandatory: bool  # If this parameter required
    expected_type: ParameterType


def is_valid_parameter_type(expected_type: ParameterType, value: Any) -> bool:
    """Returns true if the specified value "passes" as the expected type. Only performs rudimentary checks to try
    and catch obvious misconfigurations"""
    if value is None:
        return True  # We currently allow None to pass to params. Make it a runtime concern

    if is_resolvable_variable(value):
        return True  # Too hard to validate variable expressions. Make it a runtime concern

    match expected_type:
        case ParameterType.String:
            return isinstance(value, str)
        case ParameterType.Integer:
            if isinstance(value, int):
                return True
            else:
                # Floats/decimals can pass through so long as they have 0 decimal places
                try:
                    return int(value) == value
                except Exception:
                    return False
        case ParameterType.UnsignedInteger:
            # Integer that is greater than or equal to 0
            return is_valid_parameter_type(ParameterType.Integer, value) and value >= 0
        case ParameterType.Float:
            return isinstance(value, float) or isinstance(value, Decimal) or isinstance(value, int)
        case ParameterType.Boolean:
            return isinstance(value, bool)
        case ParameterType.DateTime:
            return isinstance(value, datetime)
        case ParameterType.ListString:
            return isinstance(value, list) and all((isinstance(e, str) for e in value))
        case ParameterType.HexBinary:
            try:
                int(value, 16)
                return True
            except Exception:
                return False
        case ParameterType.CSIPAusResource:
            try:
                return CSIPAusResource(value) == value
            except Exception:
                return False
        case ParameterType.ListCSIPAusResource:
            return isinstance(value, list) and all(
                (is_valid_parameter_type(ParameterType.CSIPAusResource, e) for e in value)
            )
        case ParameterType.CSIPAusReadingType:
            try:
                return CSIPAusReadingType(value) == value
            except Exception:
                return False
        case ParameterType.ListCSIPAusReadingType:
            return isinstance(value, list) and all(
                (is_valid_parameter_type(ParameterType.CSIPAusReadingType, e) for e in value)
            )
        case ParameterType.CSIPAusReadingLocation:
            try:
                return CSIPAusReadingLocation(value) == value
            except Exception:
                return False
        case ParameterType.ReadingTypeValues:
            if not value or not isinstance(value, dict):
                return False

            last_length: int | None = None
            for reading_type, reading_vals in value.items():
                if (
                    not is_valid_parameter_type(ParameterType.CSIPAusReadingType, reading_type)
                    or not isinstance(reading_vals, list)
                    or not all((is_valid_parameter_type(ParameterType.Float, rv) for rv in reading_vals))
                ):
                    return False

                if last_length is None:
                    last_length = len(reading_vals)
                elif last_length != len(reading_vals):
                    return False
            return True

    raise TestProcedureDefinitionError(f"Unexpected ParameterType: {ParameterType}")


def validate_parameters(location: str, parameters: dict[str, Any], valid_schema: dict[str, ParameterSchema]) -> None:
    """Validates parameters against valid_schema for the specified location label.

    location: Label to decorate error messages (eg TestProcedureName.Step.Action)
    parameters: The parameters dict to validate
    valid_schema: The schema to validate parameters against. Keys will be the parameter names, value will be the schema

    raises TestProcedureDefinitionError if parameters is invalid"""

    # Check the supplied parameters match the schema definition
    for param_name, param_value in parameters.items():
        param_schema = valid_schema.get(param_name, None)
        if param_schema is None:
            raise TestProcedureDefinitionError(
                f"{location} doesn't have a parameter {param_name}. Valid params are {set(valid_schema.keys())}"
            )

        # Check the type
        if not is_valid_parameter_type(param_schema.expected_type, param_value):
            raise TestProcedureDefinitionError(
                f"{location} has parameter {param_name} expecting {param_schema.expected_type} but got {param_value}"
            )

    # Check all mandatory parameters are set
    for param_name, param_schema in valid_schema.items():
        if param_schema.mandatory and param_name not in parameters:
            raise TestProcedureDefinitionError(f"{location} is missing mandatory parameter {param_name}")



from dataclasses import dataclass
from enum import StrEnum
from importlib import resources

import yaml
from cactus_test_definitions.csipaus import CSIPAusVersion
from cactus_test_definitions.schema import UniqueKeyLoader
from cactus_test_definitions.server.actions import Action
from cactus_test_definitions.server.checks import Check
from dataclass_wizard import LoadMeta, YAMLWizard


class TestProcedureId(StrEnum):
    """The set of all available test ID's

    This should be kept in sync with the current set of test procedures loaded from the procedures directory"""

    __test__ = False  # Prevent pytest from picking up this class
    S_ALL_01 = "S-ALL-01"
    S_ALL_02 = "S-ALL-02"
    S_ALL_03 = "S-ALL-03"
    S_ALL_04 = "S-ALL-04"
    S_ALL_05 = "S-ALL-05"
    S_ALL_06 = "S-ALL-06"
    S_ALL_07 = "S-ALL-07"
    S_ALL_08 = "S-ALL-08"
    S_ALL_09 = "S-ALL-09"
    S_ALL_10 = "S-ALL-10"
    S_ALL_11 = "S-ALL-11"
    S_ALL_12 = "S-ALL-12"
    S_ALL_13 = "S-ALL-13"
    S_ALL_14 = "S-ALL-14"
    S_ALL_15 = "S-ALL-15"
    S_ALL_16 = "S-ALL-16"
    S_ALL_17 = "S-ALL-17"
    S_ALL_18 = "S-ALL-18"
    S_ALL_19 = "S-ALL-19"
    S_ALL_20 = "S-ALL-20"
    S_ALL_21 = "S-ALL-21"
    S_ALL_22 = "S-ALL-22"
    S_ALL_23 = "S-ALL-23"
    S_ALL_24 = "S-ALL-24"
    S_ALL_25 = "S-ALL-25"
    S_ALL_26 = "S-ALL-26"
    S_ALL_27 = "S-ALL-27"
    S_ALL_28 = "S-ALL-28"
    S_ALL_29 = "S-ALL-29"
    S_ALL_30 = "S-ALL-30"
    S_ALL_31 = "S-ALL-31"
    S_ALL_32 = "S-ALL-32"
    S_ALL_33 = "S-ALL-33"
    S_ALL_34 = "S-ALL-34"
    S_ALL_35 = "S-ALL-35"
    S_ALL_36 = "S-ALL-36"
    S_ALL_37 = "S-ALL-37"
    S_ALL_38 = "S-ALL-38"
    S_ALL_39 = "S-ALL-39"
    S_ALL_40 = "S-ALL-40"
    S_ALL_41 = "S-ALL-41"
    S_ALL_42 = "S-ALL-42"
    S_ALL_43 = "S-ALL-43"
    S_ALL_44 = "S-ALL-44"
    S_ALL_45 = "S-ALL-45"
    # 46
    # 47
    S_ALL_48 = "S-ALL-48"
    S_ALL_49 = "S-ALL-49"
    # 50
    S_ALL_51 = "S-ALL-51"
    S_ALL_52 = "S-ALL-52"
    S_ALL_53 = "S-ALL-53"
    # 54
    S_ALL_55 = "S-ALL-55"
    S_ALL_56 = "S-ALL-56"
    S_ALL_57 = "S-ALL-57"
    S_OPT_01 = "S-OPT-01"
    S_OPT_02 = "S-OPT-02"
    S_OPT_03 = "S-OPT-03"
    S_OPT_04 = "S-OPT-04"
    S_OPT_05 = "S-OPT-05"
    # S-OPT-06 DRED


class ClientType(StrEnum):
    DEVICE = "device"  # This is a direct device client - i.e. the cert will match a SPECIFIC EndDevice
    AGGREGATOR = "aggregator"  # This is an aggregator client - i.e. the cert can manage MANY EndDevices


@dataclass
class RequiredClient:
    """A RequiredClient is a way for a test to assert that it needs a specific client type or set of clients. The id
    will be used internally within a test to reference a specific client"""

    id: str  # How this client will be referred to within the step's of the test
    client_type: ClientType | None = None  # If set - the client type that is required


@dataclass
class Step:
    """A step is an action for a client to execute and then a series of checks to validate the results. If the action
    raises an exception OR any of the checks fail, this step will marked as failed and the test will be aborted.

    Actions might represent a single operation or they may represent a series of polls/checks over a period of time.
    """

    id: str  # Descriptive identifier for this step (must be unique)
    action: Action  # The action to execute when the step starts
    client: str | None = None  # The RequiredClient.id that will execute this step. If None - use the 0th client.
    use_client_context: str | None = (
        None  # Specify to allow a request to execute with clientX using the context of clientY
    )
    checks: list[Check] | None = None  # The checks (if any) to execute AFTER action completes to determine success
    instructions: list[str] | None = None  # Text to display while this step executes

    repeat_until_pass: bool = False  # If True - failing checks will cause this step to re-execute until successful


@dataclass
class Preconditions:
    """Preconditions are a way of setting up the test / server before the test begins.

    Instructions are out-of-band information to show until the preconditions are all met.

    If any checks are required - they will be polled regularly until ALL pass. For each check poll, a discovery
    will be run to ensure data is available.
    """

    required_clients: list[RequiredClient]  # What client(s) need to be supplied to run this test procedure


@dataclass
class TestProcedure(YAMLWizard):
    """Top level object for collecting everything relevant to a single TestProcedure"""

    __test__ = False  # Prevent pytest from picking up this class
    description: str  # Metadata from test definitions
    category: str  # Metadata from test definitions
    classes: list[str]  # Metadata from test definitions
    target_versions: list[CSIPAusVersion]  # What version(s) of csip-aus is this test targeting?
    preconditions: Preconditions
    steps: list[Step]  # What behavior will the test procedure be evaluating?


LoadMeta(raise_on_unknown_json_key=True).bind_to(TestProcedure)


def parse_test_procedure(yaml_contents: str) -> TestProcedure:
    """Given a YAML string - parse a TestProcedure.

    This will ensure the YAML parser will use all the "strict" extensions to reduce the incidence of errors"""

    return TestProcedure.from_yaml(
        yaml_contents,
        decoder=yaml.load,  # type: ignore
        Loader=UniqueKeyLoader,
    )


def get_yaml_contents(test_procedure_id: TestProcedureId) -> str:
    """Finds the YAML contents for the TestProcedure with the specified TestProcedureId"""
    yaml_resource = resources.files("cactus_test_definitions.server.procedures") / f"{test_procedure_id}.yaml"
    with resources.as_file(yaml_resource) as yaml_file:
        with open(yaml_file, "r") as f:
            yaml_contents = f.read()
            return yaml_contents


def get_test_procedure(test_procedure_id: TestProcedureId) -> TestProcedure:
    """Gets the TestProcedure with the nominated ID by loading its definition from disk"""
    yaml_contents = get_yaml_contents(test_procedure_id)
    return parse_test_procedure(yaml_contents)


def get_all_test_procedures() -> dict[TestProcedureId, TestProcedure]:
    """Gets every TestProcedure, keyed by their TestProcedureId"""
    return {tp_id: get_test_procedure(tp_id) for tp_id in TestProcedureId}



from dataclasses import dataclass
from typing import Any

from cactus_test_definitions.errors import TestProcedureDefinitionError
from cactus_test_definitions.parameters import (
    ParameterSchema,
    ParameterType,
    validate_parameters,
)
from cactus_test_definitions.variable_expressions import (
    parse_variable_expression_body,
    try_extract_variable_expression,
)


@dataclass
class Action:
    type: str
    client: str | None = None  # use the client with this id to execute this action. If None, use the 0th client
    parameters: dict[str, Any] = None  # type: ignore # This will be forced in __post_init__

    def __post_init__(self):
        """Some parameter values might contain variable expressions (eg: a string "$now") that needs to be replaced
        with an parsed Expression object instead."""
        if self.parameters is None:
            self.parameters = {}

        for k, v in self.parameters.items():
            variable_expr = try_extract_variable_expression(v)
            if variable_expr:
                self.parameters[k] = parse_variable_expression_body(variable_expr, k)


# The parameter schema for each action, keyed by the action name
ACTION_PARAMETER_SCHEMA: dict[str, dict[str, ParameterSchema]] = {
    "discovery": {
        "resources": ParameterSchema(True, ParameterType.ListCSIPAusResource),  # What resources to try and resolve?
        "next_polling_window": ParameterSchema(
            False, ParameterType.Boolean
        ),  # If set - delay this until the upcoming polling window (eg- wait for the next whole minute)
        "list_limit": ParameterSchema(False, ParameterType.Integer),
    },  # Performs a full discovery / refresh of the client's context from DeviceCapability downwards
    "notifications": {
        "sub_id": ParameterSchema(True, ParameterType.String),  # Must match a previously created subscription
        "collect": ParameterSchema(
            False, ParameterType.Boolean
        ),  # Collects latest subscription notifications into context
        "disable": ParameterSchema(False, ParameterType.Boolean),  # Simulates HTTP 5XX outage at the endpoint
    },
    "wait": {
        "duration_seconds": ParameterSchema(True, ParameterType.Integer)
    },  # Waits (doing nothing - blocking other step actions) until the specified time period has passed
    "comms-status": {
        "notifications_enabled": ParameterSchema(True, ParameterType.Boolean)  # Enable/Disble notification webhook
    },  # Enables or disables certain communications
    "refresh-resource": {
        "resource": ParameterSchema(True, ParameterType.CSIPAusResource),
        "expect_rejection": ParameterSchema(False, ParameterType.Boolean),  # if set - expect 4XX and ErrorPayload
        "expect_rejection_or_empty": ParameterSchema(
            False, ParameterType.Boolean
        ),  # Similar to expect_rejection but also allow en empty list (if it's a list resource)
    },  # Force an existing resource (in the client's context) to be re-fetched via href. Updates context on success
    "insert-end-device": {
        "force_lfdi": ParameterSchema(False, ParameterType.String),  # Forces the use of this LFDI
        "expect_rejection": ParameterSchema(False, ParameterType.Boolean),  # If set - expect 4XX and ErrorPayload
    },  # Inserts an EndDevice and then validates the returned Location header
    "upsert-connection-point": {
        "connectionPointId": ParameterSchema(True, ParameterType.String),
        "expect_rejection": ParameterSchema(False, ParameterType.Boolean),  # If set - expect ErrorPayload reasonCode 1
    },
    "upsert-mup": {
        "mup_id": ParameterSchema(True, ParameterType.String),  # Used to alias the returned MUP ID
        "location": ParameterSchema(True, ParameterType.CSIPAusReadingLocation),
        "reading_types": ParameterSchema(True, ParameterType.ListCSIPAusReadingType),
        "expect_rejection": ParameterSchema(False, ParameterType.Boolean),  # If set - expect 4XX and ErrorPayload
        "mmr_mrids": ParameterSchema(
            False, ParameterType.ListString
        ),  # Must correspond 1-1 with reading_types. Used for forcing specific mrid values
        "pow10_multiplier": ParameterSchema(
            False, ParameterType.Integer
        ),  # Force the use a particular pow10. Defaults to 0 otherwise
        "set_mup_mrid": ParameterSchema(False, ParameterType.String),  # If set, forces mup mrid identity
    },  # Register a MUP with the specified values. MMR's based on hash of current client / reading types
    "insert-readings": {
        "mup_id": ParameterSchema(True, ParameterType.String),  # Must be previously defined with register-mup
        "values": ParameterSchema(
            True, ParameterType.ReadingTypeValues
        ),  # The sequences of values to send at the MUP post rate
        "expect_rejection": ParameterSchema(False, ParameterType.Boolean),  # If set - expect 4XX and ErrorPayload
    },  # Sends readings - validates that the telemetry is parsed correctly by the server
    "upsert-der-status": {
        "genConnectStatus": ParameterSchema(False, ParameterType.Integer),
        "operationalModeStatus": ParameterSchema(False, ParameterType.Integer),
        "alarmStatus": ParameterSchema(False, ParameterType.Integer),
        "expect_rejection": ParameterSchema(False, ParameterType.Boolean),  # If set - expect 4XX and ErrorPayload
    },  # Sends DERStatus - validates that the server persisted the values correctly
    "upsert-der-capability": {
        "type": ParameterSchema(True, ParameterType.Integer),
        "rtgMaxW": ParameterSchema(True, ParameterType.Integer),
        "modesSupported": ParameterSchema(True, ParameterType.Integer),
        "doeModesSupported": ParameterSchema(True, ParameterType.Integer),
    },  # Sends DERCapability - validates that the server persisted the values correctly
    "upsert-der-settings": {
        "setMaxW": ParameterSchema(True, ParameterType.Integer),
        "setGradW": ParameterSchema(True, ParameterType.Integer),
        "modesEnabled": ParameterSchema(True, ParameterType.Integer),
        "doeModesEnabled": ParameterSchema(True, ParameterType.Integer),
    },  # Sends DERSettings - validates that the server persisted the values correctly
    "send-malformed-der-settings": {
        "updatedTime_missing": ParameterSchema(True, ParameterType.Boolean),  # If true - updatedTime will be stripped
    },  # Sends a malformed DERSettings - expects a failure and that the server will NOT change anything
    "send-malformed-response": {
        "mrid_unknown": ParameterSchema(True, ParameterType.Boolean),  # If true - mrid will be random
        "endDeviceLFDI_unknown": ParameterSchema(True, ParameterType.Boolean),  # If true - endDeviceLfdi will be random
        "response_invalid": ParameterSchema(True, ParameterType.Boolean),  # If true - response will be a reserved value
    },  # Sends a malformed Response (using the most recent DERControl replyTo) - expects a failure response
    "create-subscription": {
        "sub_id": ParameterSchema(True, ParameterType.String),  # Used to alias the returned subscription ID
        "resource": ParameterSchema(True, ParameterType.CSIPAusResource),
    },  # Sends a new Subscription - validates that the server persisted the values correctly via Location
    "delete-subscription": {
        "sub_id": ParameterSchema(True, ParameterType.String),  # Must match a previously
    },  # Sends a Subscription deletion
    "respond-der-controls": {},  # Enumerates all known DERControls and sends a Response for any that require it
}
VALID_ACTION_NAMES: set[str] = set(ACTION_PARAMETER_SCHEMA.keys())


def validate_action_parameters(procedure_name: str, step_name: str, action: Action) -> None:
    """Validates the action parameters for the parent TestProcedure based on the  ACTION_PARAMETER_SCHEMA

    raises TestProcedureDefinitionError on failure"""
    location = f"{procedure_name}.step[{step_name}]"  # Descriptive location

    parameter_schema = ACTION_PARAMETER_SCHEMA.get(action.type, None)
    if parameter_schema is None:
        raise TestProcedureDefinitionError(
            f"{location} has an invalid action name '{action.type}'. Valid Names: {VALID_ACTION_NAMES}"
        )

    validate_parameters(location, action.parameters, parameter_schema)



from datetime import datetime, timedelta, timezone
from importlib import resources
from pathlib import Path

import pytest
from assertical.asserts.type import assert_dict_type
from cactus_test_definitions.client.test_procedures import (
    TestProcedure,
    TestProcedureId,
    get_all_test_procedures,
    parse_test_procedure,
)
from cactus_test_definitions.variable_expressions import (
    Constant,
    Expression,
    NamedVariable,
    NamedVariableType,
    OperationType,
)
from dataclass_wizard.errors import UnknownKeysError


def test_TestProcedureId_synchronised():
    """Ensures that TestProcedureId is in sync with all available TestProcedures YAML files.

    Each MY-TEST.yaml must have a corresponding TestProcedureId.MY_TEST"""

    # Discover all the YAML files
    suffix = ".yaml"
    raw_file_names: set[str] = set()
    for yaml_file in resources.files("cactus_test_definitions.client.procedures").iterdir():
        if yaml_file.is_file() and yaml_file.name.endswith(suffix):
            raw_file_names.add(yaml_file.name[: -len(suffix)])

    # Compare them against the TestProcedureId
    for file_name in raw_file_names:
        assert file_name in TestProcedureId
    for tp_id in TestProcedureId:
        assert tp_id in raw_file_names
    assert len(raw_file_names) == len(TestProcedureId)


def test_available_tests_populated():
    """Force test procedures to load and ensure they all validate (and we at least have a few)"""

    all_tps = get_all_test_procedures()
    assert_dict_type(TestProcedureId, TestProcedure, all_tps, count=len(TestProcedureId))
    assert all_tps[TestProcedureId.ALL_01] != all_tps[TestProcedureId.ALL_02], "Sanity check on uniqueness"


def test_error_on_duplicate_key():
    """Force test procedures to load and ensure they all validate (and we at least have a few)"""

    with open(Path("tests/data/client/tp_error_duplicate_keys.yaml"), "r") as fp:
        yaml_contents = fp.read()

    with pytest.raises(ValueError):
        parse_test_procedure(yaml_contents)


def test_error_on_extra_key():
    """Force test procedures to load and ensure they all validate (and we at least have a few)"""

    with open(Path("tests/data/client/tp_error_extra_keys.yaml"), "r") as fp:
        yaml_contents = fp.read()

    with pytest.raises(UnknownKeysError):
        parse_test_procedure(yaml_contents)


def test_TestProcedures_action_parameter_types():
    """Tests that lists/dicts/constants/datetimes can all be encoded/decoded via the yaml action definition"""

    with open(Path("tests/data/client/tp_action_parameters.yaml"), "r") as fp:
        tp = parse_test_procedure(fp.read())

    step = tp.steps["Step-1"]
    action = step.actions[0]

    assert action.parameters == {
        "param_list_str": ["List Item 1", "List Item 2", "List Item 3"],
        "param_list_int": [11, 22, 33],
        "param_list_mixed": [
            11,
            2.2,
            "List Item 3",
            datetime(2020, 1, 2, 3, 4, 5, tzinfo=timezone.utc),
        ],
        "param_dict_str": {"key1": "value1", "key2": "value2"},
        "param_dict_int": {"key11": 11, "key22": 22},
        "param_int": 11,
        "param_str": "Value 11",
        "param_datetime_utc": datetime(2025, 1, 2, 3, 4, 5, tzinfo=timezone.utc),
        "param_datetime_naive": datetime(2025, 1, 2, 3, 4, 5, tzinfo=None),
        "param_with_variable_date": NamedVariable(NamedVariableType.NOW),
        "param_with_variable_db_lookup": NamedVariable(NamedVariableType.DERSETTING_SET_MAX_W),
        "param_with_variable_relative_db_lookup": Expression(
            OperationType.MULTIPLY, Constant(0.25), NamedVariable(NamedVariableType.DERSETTING_SET_MAX_W)
        ),
        "param_with_variable_relative_date": Expression(
            OperationType.SUBTRACT, NamedVariable(NamedVariableType.NOW), Constant(timedelta(minutes=5))
        ),
    }



from cactus_test_definitions.errors import TestProcedureDefinitionError
from cactus_test_definitions.server.actions import validate_action_parameters
from cactus_test_definitions.server.checks import validate_check_parameters
from cactus_test_definitions.server.test_procedures import (
    TestProcedure,
    TestProcedureId,
)


def validate_test_procedure(test_procedure: TestProcedure, test_procedure_id: TestProcedureId):
    # Check preconditions
    if not test_procedure.preconditions.required_clients:
        raise TestProcedureDefinitionError(
            f"{test_procedure_id} has no RequiredClients element. At least 1 entry required"
        )
    required_clients_by_id = dict(((rc.id, rc) for rc in test_procedure.preconditions.required_clients))

    for step in test_procedure.steps:
        validate_action_parameters(test_procedure_id, step.id, step.action)

        # Validate step checks
        if step.checks:
            for check in step.checks:
                validate_check_parameters(test_procedure_id, check)

        # Ensure client exists
        if step.client is not None and step.client not in required_clients_by_id:
            raise TestProcedureDefinitionError(
                f"{test_procedure_id} reference client {step.client} that isn't listed in RequiredClients."
            )



from cactus_test_definitions.client.actions import Action, validate_action_parameters
from cactus_test_definitions.client.checks import validate_check_parameters
from cactus_test_definitions.client.events import validate_event_parameters
from cactus_test_definitions.client.test_procedures import (
    TestProcedure,
    TestProcedureId,
)
from cactus_test_definitions.errors import TestProcedureDefinitionError


def validate_action(
    procedure: TestProcedure, test_procedure_id: TestProcedureId, location: str, action: Action
) -> None:
    """Handles the full validation of an action's definition for a parent procedure.

    procedure: The parent TestProcedure for action
    test_procedure_id: The name of procedure (used for labelling errors)
    location: Where in procedure can you find action? (used for labelling errors)
    action: The action to validate

    raises TestProcedureDefinitionError on failure
    """
    validate_action_parameters(test_procedure_id, location, action)

    # Provide additional "action specific" validation
    match action.type:
        case "enable-steps" | "remove-steps":
            for step_name in action.parameters["steps"]:
                if step_name not in procedure.steps.keys():
                    raise TestProcedureDefinitionError(
                        f"{test_procedure_id}.{location}. Refers to unknown step '{step_name}'."
                    )


def validate_test_procedure_actions(test_procedure: TestProcedure, test_procedure_id: TestProcedureId) -> None:
    """Validate actions of test procedure steps / preconditions

    Ensure,
    - action has the correct parameters
    - if parameters refer to steps then those steps are defined for the test procedure
    """

    # Validate actions in the preconditions
    if test_procedure.preconditions:
        if test_procedure.preconditions.actions:
            for action in test_procedure.preconditions.actions:
                validate_action(test_procedure, test_procedure_id, "Precondition", action)

        if test_procedure.preconditions.init_actions:
            for action in test_procedure.preconditions.init_actions:
                validate_action(test_procedure, test_procedure_id, "Precondition", action)

    # Validate actions that exist on steps
    for step_name, step in test_procedure.steps.items():
        for action in step.actions:
            validate_action(test_procedure, test_procedure_id, step_name, action)


def validate_test_procedure_checks(test_procedure: TestProcedure, test_procedure_id: TestProcedureId) -> None:
    """Validate checks of a test procedure

    Ensure,
    - check has the correct parameters
    """

    if test_procedure.criteria and test_procedure.criteria.checks:
        for check in test_procedure.criteria.checks:
            validate_check_parameters(f"{test_procedure_id}: Criteria", check)

    if test_procedure.preconditions and test_procedure.preconditions.checks:
        for check in test_procedure.preconditions.checks:
            validate_check_parameters(f"{test_procedure_id}: Preconditions", check)

    for step_name, step in test_procedure.steps.items():
        if step.event.checks:
            for check in step.event.checks:
                validate_check_parameters(f"{test_procedure_id}: Step {step_name}", check)


def validate_test_procedure_events(test_procedure: TestProcedure, test_procedure_id: TestProcedureId) -> None:
    """Validate events of test procedure steps

    Ensure,
    - event has the correct parameters
    """

    if test_procedure.steps:
        for step_name, step in test_procedure.steps.items():
            validate_event_parameters(test_procedure_id, step_name, step.event)


def validate_test_procedure(test_procedure: TestProcedure, test_procedure_id: TestProcedureId) -> None:
    """Performs additional "high level" validation of a test procedure. (eg: ensuring all action names are valid)

    raises TestProcedureDefinitionError on error"""
    validate_test_procedure_actions(test_procedure, test_procedure_id)
    validate_test_procedure_checks(test_procedure, test_procedure_id)
    validate_test_procedure_events(test_procedure, test_procedure_id)



from importlib import resources
from pathlib import Path

import pytest
from assertical.asserts.type import assert_dict_type
from cactus_test_definitions.server.test_procedures import (
    TestProcedure,
    TestProcedureId,
    get_all_test_procedures,
    parse_test_procedure,
)
from dataclass_wizard.errors import UnknownKeysError


def test_TestProcedureId_synchronised():
    """Ensures that TestProcedureId is in sync with all available TestProcedures YAML files.

    Each MY-TEST.yaml must have a corresponding TestProcedureId.MY_TEST"""

    # Discover all the YAML files
    suffix = ".yaml"
    raw_file_names: set[str] = set()
    for yaml_file in resources.files("cactus_test_definitions.server.procedures").iterdir():
        if yaml_file.is_file() and yaml_file.name.endswith(suffix):
            raw_file_names.add(yaml_file.name[: -len(suffix)])

    # Compare them against the TestProcedureId
    for file_name in raw_file_names:
        assert file_name in TestProcedureId
    for tp_id in TestProcedureId:
        assert tp_id in raw_file_names
    assert len(raw_file_names) == len(TestProcedureId)


def test_available_tests_populated():
    """Force test procedures to load and ensure they all validate (and we at least have a few)"""

    all_tps = get_all_test_procedures()
    assert_dict_type(TestProcedureId, TestProcedure, all_tps, count=len(TestProcedureId))
    assert all_tps[TestProcedureId.S_ALL_01] != all_tps[TestProcedureId.S_ALL_02], "Sanity check on uniqueness"


def test_error_on_duplicate_key():
    """Force test procedures to load and ensure they all validate (and we at least have a few)"""

    with open(Path("tests/data/server/tp_error_duplicate_keys.yaml"), "r") as fp:
        yaml_contents = fp.read()

    with pytest.raises(ValueError):
        parse_test_procedure(yaml_contents)


def test_error_on_extra_key():
    """Force test procedures to load and ensure they all validate (and we at least have a few)"""

    with open(Path("tests/data/server/tp_error_extra_keys.yaml"), "r") as fp:
        yaml_contents = fp.read()

    with pytest.raises(UnknownKeysError):
        parse_test_procedure(yaml_contents)
